# Authors : Guillaume Marques
# Description : Procedure to separate SSI

# Entry function
function separate_ssi!(vic::SSIContainer, λ_vals::Vector{Double}, 
    y_vals::Array{Double, 3}, C::Vector{Int}, S::Vector{Int}, Q1::Int, Q2::Int, 
    g2l::TierGraph, cust_demands::Vector{Double}, sats_supply::Vector{Double}
    )::Cvoid
  ssd = SSISeparationData(C, S, edges(g2l), Q1, Q2, cust_demands, sats_supply, 
      λ_vals, y_vals)
  separate_ssi!(vic, ssd)
  return
end

function separate_ssi!(vic::SSIContainer, ssd::SSISeparationData)::Cvoid
  # Step 2 : Satellites filtering
  subsetstoseparate = satsubsetsfilter(length(ssd.satellites), ssd.λ_vals)
  # Step 3 : Separate SSI for each subset
  for sss in subsetstoseparate
    sep_time = @elapsed separate_ssi_for_satsubset!(vic, ssd, sss)
    flush_buffer!(vic, sep_time)
  end
  return
end

function in_out_undecided_sets(ssd::SSISeparationData
    )::Tuple{Vector{Int}, Vector{Int}, Vector{Int}}
  in_set = Vector{Int}() # Nodes ∈ (C ∪ S)
  out_set = Vector{Int}() # Nodes ∉ (C ∪ S)
  undecided_set = Vector{Int}() # Nodes that can ∈ or ∉ (C ∪ S)
  ϵ = 1e-6
  for i in ssd.customers
    incident_edges_val = incident_edges_w(ssd, i)
    if 2 - ϵ <= incident_edges_val
      # The node is already visited by satellites ∉ S.
      # The node ∉ C
      push!(out_set, i)
    elseif incident_edges_val <= ϵ
      # The node is only visited by satellites ∈ S.
      # If there is not enough freight at satellites ∈ S, the node will be
      # supplied by satellites ∉ S. The node ∈ C.
      push!(in_set, i)
    else
      push!(undecided_set, i)
    end
  end
  for s in ssd.satellites[.!dep_sats(ssd)]
    push!(out_set, s)
  end
  return in_set, out_set, undecided_set
end

ord(i, j) = (i < j) ? (i, j) : (j, i)
function check_ssi_violation!(vic::SSIContainer, ssd::SSISeparationData,
    vertices_covered::BitArray, vertices_uncovered::BitArray, 
    cutval::Double)::Bool
  if !(reduce(|, vertices_covered)) 
    #@warn("Set of covered vertices is empty.")
    return false
  end
  violated_ineq_found = false
  demand_covered = sum(ssd.demands[vertices_covered])
  breakpoints = compute_breakpoints(ssd, demand_covered)
  for i in 1:(length(breakpoints) - 1)
    b = b_(ssd, demand_covered, breakpoints[i], breakpoints[i+1])
    m = m_(ssd, demand_covered, breakpoints[i], breakpoints[i+1])
    computed_lhs = cutval - 2 * m * nburbantrucks(ssd)
    rhs = 2 * b
    #info("\e[32m Try : \e[0m  demand    = \e[1m $demand_covered \e[0m  -  m = \e[1m $m \e[0m  - lhs = \e[1m $computed_lhs \e[0m -  rhs = \e[1m $rhs \e[0m.  (Ω[$i] = $(Ω[i]) & Ω[$(i+1)] = $(Ω[i+1]))")
    violation = rhs - computed_lhs
    if violation > 1e-6
      # info("Found violated constraint lhs = $computed_lhs - rhs = $rhs")
      # info("\e[35m Registering violation : \e[0m  demand = \e[1m $demand_covered \e[0m  -  m = \e[1m $m \e[0m  - lhs = \e[1m $computed_lhs \e[0m -  rhs = \e[1m $rhs \e[0m.  (Ω[$i] = $(breakpoints[i]) & Ω[$(i+1)] = $(breakpoints[i+1]))")
      ineq = SatellitesSupplyInequality(m, rhs, copy(vertices_covered), 
          copy(vertices_uncovered), satsubsetid(ssd), violation)
      add_violated_inequality!(vic, ineq)
      violated_ineq_found = true
      ## Following code to debug
      # real_lhs = real_lhs2 = - 2 * m * nburbantrucks(ssd)
      #   for i in ssd.vertices[vertices_covered]
      #    for j in ssd.vertices[vertices_uncovered]
      #      real_lhs += edge_w(ssd, i, j)
      #      real_lhs2 += sum(getvalue(ssd.y[s, ord(i, j)]) for s in ssd.satellites[.!dep_sats(ssd)])
      #    end
      #   end
      # @printf "Checking lhs . computed_lhs = %0.5f \n" computed_lhs
      # @printf "Checking lhs ..... real_lhs = %0.5f \n" real_lhs
      # @printf "Checking lhs2 .... real_lhs = %0.5f \n" real_lhs2
    end
  end
  return violated_ineq_found
end

function get_covered_and_uncovered_vertices(ssd::SSISeparationData,
    vnm::VertexNodeMaps, partition::Vector{Int8})::Tuple{BitArray, BitArray}
  # Last node is in "out" set
  uncovered_id = partition[end]
  vertices_covered = falses(ssd.n)
  vertices_uncovered = falses(ssd.n)
  for (n, p) in enumerate(partition)
    if p == uncovered_id
      for v in node_to_vertices(vnm, n)
        vertices_uncovered[v] = true
      end
    else
      for v in node_to_vertices(vnm, n)
        vertices_covered[v] = true
      end
    end
  end
  for v in vnm.in_vertices
    vertices_covered[v] = true
  end
  return vertices_covered, vertices_uncovered
end

function update_cov_uncov_partitions!(vnm::VertexNodeMaps, 
    undecided_cov::BitArray, v_cov::BitArray, v_uncov::BitArray)::Cvoid
  for (i, c) in enumerate(vnm.undecided_vertices)
    v_cov[c] = undecided_cov[i]
    v_uncov[c] = !undecided_cov[i]
  end
  return
end

function separate_ssi_for_satsubset!(vic::SSIContainer, 
    ssd::SSISeparationData, subsetid::Int)::Cvoid
  # Update data to separate SSI for the given satellites subset
  updateSSIdata!(ssd, subsetid)
  # Create the graph and the map
  vnm, g, distmx = creategraph(ssd)
  (nv(g) == 1) && return
  allnodespartitioned(vnm) && return
  # Minimum cut
  partition, cutval = mincut(g, distmx)
  v_cov, v_uncov = get_covered_and_uncovered_vertices(ssd, vnm, partition)
  violation_with_mincut = check_ssi_violation!(vic, ssd, v_cov, v_uncov, cutval)
  greedy_to_find_violated_ssi!(vic, ssd, vnm, copy(v_cov), copy(v_uncov))
  #heuristic_to_find_violated_ssi!(vic, ssd, vnm, copy(v_cov), copy(v_uncov))
  return
end

function greedy_to_find_violated_ssi!(vic::SSIContainer,
    ssd::SSISeparationData, vnm::VertexNodeMaps, v_cov::BitArray, 
    v_uncov::BitArray)::Cvoid
  # We only work on "undecided" set of customers
  undecided_cov = v_cov[vnm.undecided_vertices]
  ratio_dw = zeros(Double, length(undecided_cov))
  while !reduce(&, undecided_cov) && minimum(ratio_dw) < Inf
    move_from_uncov_to_cov!(ssd, vnm, undecided_cov, v_uncov, ratio_dw)
    update_cov_uncov_partitions!(vnm, undecided_cov, v_cov, v_uncov)
    cutval = compute_cut_value(ssd, v_cov, v_uncov)
    check_ssi_violation!(vic, ssd, v_cov, v_uncov, cutval)
  end
  return
end

function heuristic_to_find_violated_ssi!(vic::SSIContainer, 
    ssd::SSISeparationData, vnm::VertexNodeMaps, v_cov::BitArray,
    v_uncov::BitArray)::Cvoid
  # We only work on "undecided" set of customers
  undecided_cov = v_cov[vnm.undecided_vertices]
  ratio_dw = zeros(Double, length(undecided_cov))
  #### Parameters of the heuristic
  max_iterations = 2 * length(undecided_cov)
  prob_add_cust = 0.7
  prob_random_pick_from_uncov = 0.3
  prob_random_pick_from_cov = 0.5
  #tabu_list_size = Int(ceil(0.2 * length(undecided_cov)))
  #####
  #tabu_list = Queue(Int)
  i = 0
  while i < max_iterations
    all_covered = reduce(&, undecided_cov)
    all_uncovered = !reduce(|, undecided_cov)
    change = true
    if all_covered
      change = swap_randomly!(undecided_cov, 0.75)
    elseif all_uncovered
      change = swap_randomly!(undecided_cov, 0.25)
    else
      r_op = rand(rng)
      r_meth = rand(rng)
      if r_op <= prob_add_cust
        if r_meth <= prob_random_pick_from_uncov
          change = move_randomly!(undecided_cov, false, 0.3)
        else
          move_from_uncov_to_cov!(ssd, vnm, undecided_cov, v_uncov, ratio_dw)
        end
      else
        if r_meth <= prob_random_pick_from_cov
          change = move_randomly!(undecided_cov, true, 0.2)
        else
          move_from_cov_to_uncov!(ssd, vnm, undecided_cov, v_cov, ratio_dw)
        end
      end
    end
    if change
      update_cov_uncov_partitions!(vnm, undecided_cov, v_cov, v_uncov)
      cutval = compute_cut_value(ssd, v_cov, v_uncov)
      check_ssi_violation!(vic, ssd, v_cov, v_uncov, cutval)
    end
    i += 1
  end
  return
end

function ratio_resdem_w(ssd::SSISeparationData, c::Int, target_set::BitArray)::Double
  return residual_demand(ssd, c) / (incident_edges_w(ssd, c, target_set) + 0.0000001)
end

function update_ratio_resdem_weight!(vec::Vector{Double}, 
  ssd::SSISeparationData, vnm::VertexNodeMaps, target_set::BitArray)::Cvoid
  map!(v -> 0, vec, vec)
  map!(c -> ratio_resdem_w(ssd, c, target_set), vec, vnm.undecided_vertices)
  return
end

function swap_randomly!(set::BitArray, prob_swap::Double)::Bool
  change = false
  for i in 1:length(set)
    p = rand(rng)
    if p <= prob_swap
      set[i] = !set[i]
      change = true
    end
  end
  return change
end

function move_randomly!(set::BitArray, select_val::Bool, prob_move::Double)::Bool
  change = false
  for i in 1:length(set)
    if set[i] == select_val
      p = rand(rng)
      if p <= prob_move
        set[i] = !select_val
        change = true
      end
    end
  end
  return change
end

function move_from_uncov_to_cov!(ssd::SSISeparationData, vnm::VertexNodeMaps, 
    undecided_cov::BitArray, v_uncov::BitArray, r::Vector{Double})::Cvoid
  update_ratio_resdem_weight!(r, ssd, vnm, v_uncov)
  swap_vertex = argmax(r .* (-1 .* undecided_cov .+ 1))
  undecided_cov[swap_vertex] = true
  return
end

function move_from_cov_to_uncov!(ssd::SSISeparationData, vnm::VertexNodeMaps, 
    undecided_cov::BitArray, v_cov::BitArray, r::Vector{Double})::Cvoid
  update_ratio_resdem_weight!(r, ssd, vnm, v_cov)
  swap_vertex = argmin(r .* (Inf .* .!(undecided_cov) .+ 1))
  undecided_cov[swap_vertex] = false
  return
end

function compute_cut_value(ssd::SSISeparationData, v_cov::BitArray, 
    v_uncov::BitArray)::Double
  cut_value = 0.0
  for i in ssd.vertices[v_cov], j in ssd.vertices[v_uncov]
      cut_value += edge_w(ssd, i, j)
  end
  return cut_value
end
