# Author : Guillaume Marques
# Description : Datastructures for SSI separation

# A structure to map input vertices and nodes used in the algorithm.
struct VertexNodeMaps
  undecided_vertices::Array{Int}
  in_vertices::Array{Int}
  out_vertices::Array{Int} # in uncov set (always something inside)
  vertices_map::Array{Int} # vertex => node
  nodes_map::Array{Array{Int}} # node => vertex/vertices
end

function VertexNodeMaps(ssd::SSISeparationData, undecided_vertices::Vector{Int}, in_vertices::Vector{Int}, out_vertices::Vector{Int})
  # nodes -> vertex
  nodes_map = Vector{Vector{Int}}()
  for u in undecided_vertices
    push!(nodes_map, [u])
  end
  push!(nodes_map, out_vertices)
  # vertex -> nodes
  vertices_map = zeros(Int, ssd.n)
  for (i, nodes) in enumerate(nodes_map)
    for node in nodes
      vertices_map[node] = i
    end
  end
  return VertexNodeMaps(undecided_vertices, in_vertices, out_vertices,
      vertices_map, nodes_map)
end

function infoabout(vnm::VertexNodeMaps)::Cvoid
  println("\e[1m info about VertexNodeMaps. \e[0m")
  println(" > vertices covered : $(vnm.in_vertices).")
  println(" > vertices uncovered : $(vnm.out_vertices).")
  return
end

vertex_to_node(vnm::VertexNodeMaps, v)::Int = vnm.vertices_map[v]
node_to_vertices(vnm::VertexNodeMaps, n)::Vector{Int} = vnm.nodes_map[n]
function allnodespartitioned(vnm::VertexNodeMaps)::Bool
  return (length(vnm.in_vertices) + length(vnm.out_vertices)) == length(vnm.nodes_map)
end

function demand_covered(ssd::SSISeparationData, vnm::VertexNodeMaps)
  demand_val = 0
  for v in vnm.in_vertices
    demand_val += ssd.demands[v]
  end
  return demand_val
end

function demand(ssd::SSISeparationData, vnm::VertexNodeMaps, n::Int)
  vertices = node_to_vertices(vnm, n)
  demand_val = 0
  for v in vertices
    demand_val += ssd.demands[v]
  end
  return demand_val
end

function creategraph(ssd::SSISeparationData)::Tuple{VertexNodeMaps, Graph, Matrix{Double}}
  # Nodes in "in" do not have neighbors : we will delete them.
  # Nodes in "out" cannot be in the coevered set : we will merge them
  # We will work with nodes in undecided
  in, out, undecided = in_out_undecided_sets(ssd)
  # Create nodes <-> vertices maps
  vnm = VertexNodeMaps(ssd, undecided, in, out)
  # Create the graph
  n = length(undecided) + 1 # 1 is for node corresponding to "out" merged 
  (n != length(vnm.nodes_map)) && error("Length of nodes map not equals to the number of nodes.")
  # Register nodes
  g = Graph(n)
  # Register edges & create distance matrix
  distmx = zeros(Double, n, n)
  for i in 1:(n - 1)
    for j in (i + 1):n
      w = 0
      for k in node_to_vertices(vnm, i), l in node_to_vertices(vnm, j)
        w += edge_w(ssd, k, l)
      end
      if w > 0
        add_edge!(g, (i, j))
        distmx[i, j] = distmx[j, i] = w
      end
    end
  end
  return vnm, g, distmx
end
