# Author : Guillaume Marques
# Description : Data structures to separate SSI


# data specific to the separation of SSI for a given subset of satellites
mutable struct SSISatSubsetData
  satsubset::BitArray # true if satellite in S
  satsubsetid::Int
  edges_weight::Matrix{Double}
  residual_demands::Vector{Double}
  first_level_routes::BitArray # first lev routes visiting the subset
  nb_urban_trucks::Double
end

function SSISatSubsetData(nbsatellites::Int, nbcustomers::Int)
  nbvertices = nbsatellites + nbcustomers
  return SSISatSubsetData(falses(nbsatellites), 0, 
    zeros(Double, nbvertices, nbvertices), zeros(Double, nbcustomers), 
    falses(2^nbsatellites - 1), 0.0)
end

# data for the separation of SSI
mutable struct SSISeparationData
  n::Int # nb of nodes
  vertices::Vector{Int} # id of vertices (avoid allocating array)
  first_level_routes::Vector{Int} 
  second_level_edges::Vector{Tuple} # Edges of 2nd level graph
  customers::Vector{Int}
  satellites::Vector{Int}
  Q1::Int
  Q2::Int
  demands::Vector{Double} # Concatenation of customers and satellites demands
  customers_visitation::Matrix{Double}
  sats_supply::Vector{Double}
  λ_vals::Vector{Double}
  y_vals::Array{Double, 3} #[ y_[s, i, j] for s in S, for (i, j) in E2]
  data_satsubset::SSISatSubsetData
end

sssdata(ssd::SSISeparationData)::SSISatSubsetData = ssd.data_satsubset
dep_sats(ssd::SSISeparationData)::BitArray = sssdata(ssd).satsubset
satsubsetid(ssd::SSISeparationData) = sssdata(ssd).satsubsetid
nburbantrucks(ssd::SSISeparationData) = sssdata(ssd).nb_urban_trucks
function residual_demand(ssd::SSISeparationData, c::Int)::Double
  return sssdata(ssd).residual_demands[c]
end
function residual_demands(ssd::SSISeparationData)::Vector{Double}
  return sssdata(ssd).residual_demands
end

function SSISeparationData(customers::Vector{Int}, satellites::Vector{Int},
    edges_2l::Vector{Tuple}, Q1::Int, Q2::Int, cust_demands::Vector{Double}, 
    sats_supply::Vector{Double}, λ_vals::Vector{Double}, 
    y_vals::Array{Double, 3})
  nb_customers = length(customers)
  nb_satellites = length(satellites)
  nb_vertices = nb_customers + nb_satellites
  nb_first_level_routes = 2^nb_satellites - 1
  demands = vcat(cust_demands, zeros(Double, nb_satellites))
  customers_visitation = zeros(Double, nb_customers, nb_satellites)
  offset = satellites[1] - 1
  for s in satellites .- offset
    for j in 1:nb_vertices
      for i in 1:j
        w = y_vals[i, j, s]
        if i <= nb_customers
          customers_visitation[i, s] += w / 2
        end
        if j <= nb_customers
          customers_visitation[j, s] += w / 2
        end
      end
    end
  end
  return SSISeparationData(nb_vertices, collect(1:nb_vertices), 
      collect(1:nb_first_level_routes), edges_2l, customers, satellites, Q1, Q2, 
      demands, customers_visitation, sats_supply, λ_vals, y_vals, 
      SSISatSubsetData(nb_satellites, nb_customers))
end

function updateSSIdata!(ssd::SSISeparationData, satsubsetid::Int)::Cvoid
  nbsats = length(ssd.satellites)
  nbcusts = length(ssd.customers)
  update_satsubset!(ssd.data_satsubset, nbsats, satsubsetid)
  update_edges_weights!(ssd.data_satsubset, ssd)
  update_residual_demands!(ssd.data_satsubset, ssd, satsubsetid)
  update_routes_visiting!(ssd.data_satsubset, nbsats, satsubsetid)
  nb_ut = sum(ssd.λ_vals[ssd.data_satsubset.first_level_routes])
  ssd.data_satsubset.satsubsetid = satsubsetid
  ssd.data_satsubset.nb_urban_trucks = nb_ut
  return
end

function update_edges_weights!(ssidata::SSISatSubsetData, 
    ssd::SSISeparationData)::Cvoid
  fill!(ssidata.edges_weight, 0)
  nbvertices = ssd.n
  offset = ssd.satellites[1] - 1
  for s in ssd.satellites[.!dep_sats(ssd)] .- offset
    for j in 1:nbvertices
      for i in 1:j
        w = ssd.y_vals[i, j, s]
        ssidata.edges_weight[i, j] += w
        ssidata.edges_weight[j, i] += w
      end
    end
  end
  return
end

function update_residual_demands!(ssidata::SSISatSubsetData, 
    ssd::SSISeparationData, satsubsetid::Int)::Cvoid
  fill!(ssidata.residual_demands, 0)
  nbcusts = length(ssd.customers)
  for c in 1:nbcusts
    c_visit = sum(ssd.customers_visitation[c, .!dep_sats(ssd)])
    if c_visit > 1 # Fix ϵ-gap
      c_visit = 1
    end
    ssidata.residual_demands[c] = (1 - c_visit) * ssd.demands[c]
  end
  return
end

function update_routes_visiting!(ssidata::SSISatSubsetData, nbsats::Int, 
    satsubsetid::Int)::Cvoid
  nbroutes = 2^nbsats - 1
  for p in 1:nbroutes
    ssidata.first_level_routes[p] = (p & satsubsetid > 0)
  end
  return
end

function update_satsubset!(ssidata::SSISatSubsetData, nbsats::Int, 
    satsubsetid::Int)::Cvoid
  for s in 1:nbsats
    ssidata.satsubset[s] = (2^(s - 1) & satsubsetid > 0) 
  end
  return
end

function compute_residual_demands(ssd::SSISeparationData, 
    satsubset::Vector{Int})::Vector{Double}
  nbcusts = length(ssd.customers)
  residual_demand = zeros(Double, nbcusts)
  offset = ssd.satellites[1] - 1
  for c in 1:nbcusts
    c_visit = sum(ssd.customers_visitation[c, s - offset] for s in satsubset)
    if c_visit > 1 # Fix ϵ-gap
      c_visit = 1
    end
    residual_demand[c] = (1 - c_visit) * ssd.demands[c]
  end
  return residual_demand
end

function closed_satellites(ssd::SSISeparationData)::Vector{Int}
  closed_sats = Vector{Int}()
  for s in ssd.satellites
    id_vector_s = s - ssd.satellites[1] + 1
    if ssd.sats_supply[id_vector_s] == 0
      push!(closed_sats, s)
    end
  end
  return closed_sats
end

function satellites_without_residual_demand(ssd::SSISeparationData)::Vector{Int}
  sats_no_res_dem = Vector{Int}()
  for s in ssd.satellites
    id_matrix_s = s - ssd.satellites[1] + 1
    visitations = sum(ssd.customers_visitation[:, id_matrix_s])
    supplied_custs = sum(ceil.(ssd.customers_visitation[:, id_matrix_s]))
    if visitations - supplied_custs >= 0 # should equal to 0
      push!(sats_no_res_dem, s)
    end
  end
  return sats_no_res_dem
end

function k_(ssd::SSISeparationData, cov_dem::Double)::Int 
  return Int(ceil(cov_dem / ssd.Q1))
end

function r_(ssd::SSISeparationData, cov_dem::Double, i::Int)::Int
  return Int(max(0, ceil((cov_dem - i * ssd.Q1) / ssd.Q2)))
end

function m_(ssd::SSISeparationData, cov_dem::Double, i1::Int, i2::Int)::Double
  return (r_(ssd, cov_dem, i2) - r_(ssd, cov_dem, i1)) / (i2 - i1)
end

function b_(ssd::SSISeparationData, cov_dem::Double, i1::Int, i2::Int)::Double
  return  (i2 * r_(ssd, cov_dem, i1) - i1 * r_(ssd, cov_dem, i2)) / (i2 - i1)
end

nodes(ssd::SSISeparationData)::Int = ssd.n

function compute_breakpoints(ssd::SSISeparationData, demand_covered::Double)::Vector{Int}
  k = k_(ssd, demand_covered)
  Ω = Vector{Int}()
  push!(Ω, 0)
  w = 0
  prev_m = 99999
  while w < k
    min_ww = 99999
    min_m = 99999.9
    for ww in (w+1):k
      m = m_(ssd, demand_covered, Ω[end], ww)
      if m <= min_m && ((length(Ω) == 1) || (m >= prev_m))
        min_m = m
        min_ww = ww
      end
    end
    push!(Ω, min_ww)
    w = min_ww
    prev_m = min_m
  end
  return Ω
end

function edge_w(ssd::SSISeparationData, i::Int, j::Int)::Double
  return ssd.data_satsubset.edges_weight[i, j]
end

function incident_edges_w(ssd::SSISeparationData, v::Int)::Double
  return sum(ssd.data_satsubset.edges_weight[v, :])
end

function incident_edges_w(ssd::SSISeparationData, v::Int, 
    only_vert::BitArray)::Double
  return sum(ssd.data_satsubset.edges_weight[v, only_vert])
end

function totaldemand(ssd::SSISeparationData)::Double
  return sum(ssd.demands)
end
