# Authors : Guillaume Marques
# Description : Bellman Held Karp based algorithm to enumerate 1st level routes

function subsets_length(nbsubsets::Int)::Vector{Vector{Int}}
  max_size = count_ones(nbsubsets)
  sss_sizes = [Vector{Int}() for i in 1:max_size]
  for i in 1:nbsubsets
    push!(sss_sizes[count_ones(i)], i)
  end
  return sss_sizes
end

function onespos(sss_number::Int)::Vector{Int}
  bin_sats = reverse(string(sss_number, base = 2))
  pos = Vector{Int}()
  for i in 1:length(bin_sats)
    if bin_sats[i] == '1'
      push!(pos, i)
    end
  end
  return pos
end

function number_to_satellites(sss_number::Int64, g::TierGraph)::Vector{Int}
  return onespos(sss_number) .+ offset(g)
end

function route_departure(data::Data2EVRP, p::Int)::Int
  D = cdc(data)
  nbsats = nbsatellites(data)
  nbsubsets = 2^nbsats - 1
  depot = D[Int(ceil(p / nbsubsets))]
  return depot
end

function nodes_visited_by_route(data::Data2EVRP, g::TierGraph, p::Int)::Vector{Int}
  nbsats = nbsatellites(data)
  nbsubsets = 2^nbsats - 1
  depot_offset = Int(floor((p - 1) / nbsubsets))
  nodes = number_to_satellites(p - depot_offset * nbsubsets, g)
  depot = route_departure(data, p)
  push!(nodes, depot)
  return nodes
end

function heldkarp_tsp_enum(data::Data2EVRP, g::TierGraph)::Vector{Double}
  D = cdc(data)
  nbcdc = length(D)
  nbsats = nbsatellites(data)
  nbsubsets = 2^nbsats - 1
  tours_cost = [Inf for i in 1:(nbsubsets * nbcdc)]
  sss_of_size = subsets_length(nbsubsets)
  max_size = count_ones(nbsubsets)
  tours_id_offset = 0
  for depot in D 
    labels_cost = [Inf for s in 1:nbsubsets, l in 1:nbsats]
    cur_size = 1
    while cur_size <= max_size
      for sss in sss_of_size[cur_size]
        S = number_to_satellites(sss, g)
        if cur_size == 1
          tours_cost[sss + tours_id_offset] = 2 * nodes_distance(data, depot, S[1])
          labels_cost[sss, S[1] - offset(g)] = nodes_distance(data, depot, S[1])
        else
          for i in S
              # Compute C(S, i)
              nsss = sss - 2^(i - 1 - offset(g))
              minlabelcost = minimum([labels_cost[nsss, j - offset(g)] + nodes_distance(data, j, i) for j in setdiff(S, i)])
              labels_cost[sss, i - offset(g)] = minlabelcost
              tours_cost[sss + tours_id_offset] = min(tours_cost[sss], minlabelcost + nodes_distance(data, i, depot))
          end
        end
      end
      cur_size += 1
    end
    tours_id_offset += nbsubsets
  end
  return tours_cost
end

function route_visits_node(data::Data2EVRP, g::TierGraph, p::Integer, s::Integer)
  nbcust = nbcustomers(data)
  nbsats = nbsatellites(data)
  nbsubsets = 2^nbsats - 1
  dep_offset = Int(floor((p - 1) / nbsubsets))
  nodesvisited = p - dep_offset * nbsubsets
  if s <= nbcust # this is a customer
    return 0
  elseif s > nbcust + nbsats # this is a depot
    return (s == nbcust + nbsats + dep_offset + 1) ? 1 : 0
  else # this is a satellite
    return ((nodesvisited & 2^(s - offset(g) - 1)) > 0) ? 1 : 0
  end
end

function route_visits_node(satellites::Vector{Int}, p::Integer, s::Integer)::Int
  return (p & 2^(s - satellites[1]) > 0) ? 1 : 0
end

function routes_visiting_all(data::Data2EVRP, subsetid::Int)::Vector{Int}
  routes = Vector{Int}()
  nbsats = nbsatellites(data)
  nbroutes = 2^nbsats - 1
  for p in 1:nbroutes
    if p & subsetid == subsetid
      push!(routes, p)
    end
  end
  return routes
end

function routes_visiting(data::Data2EVRP, subsetid::Int)::BitArray
  return routes_visiting(nbsatellites(data), subsetid)
end

function routes_visiting(nbsats::Int, subsetid::Int)::BitArray
  nbroutes = 2^(nbsats) - 1
  routes = falses(nbroutes)
  for p in 1:nbroutes
    routes[p] = (p & subsetid > 0)
  end
  return routes
end
