## Solution checker
function check_solution(optimizer,p::Problem)
  data = getdata(p)
  variables = getvariables(p)
  println("\e[1;33m Retrieve solution for $(typeof(variables)). \e[00m")
  solution = Solution(optimizer, data, variables)
  return check_solution(optimizer, data, solution)
end

# Assert customers id begins @ 1
function check_solution(optimizer, data::Data2EVRP, s::SolutionCapacitated)::Bool
  sol_cost = 0
  feasible_solution = true

  K = veh_1l(data)[1]
  L = veh_2l(data)[1]

  # Print data from first level
  ## City distribution center
  println("\e[1;34m City distribution centers used.\e[0m")
  for (d, items_transfered) in s.items_shipped_from_cdc
    sol_cost += (items_transfered > 0) ? node_cost(data, d) : 0
    feasible_solution &= (items_transfered <= node_capacity(data, d))
    !feasible_solution && @warn("Items transfer infeasible.")
    println(" > cdc \e[1m$d\e[0m shipping \e[1m$items_transfered\e[0m items  (setup_cost = $(node_cost(data, d)) & cap = $(node_capacity(data, d))).")
  end

  ## First level routes
  println("\e[1;34m First level routes used.\e[0m")
  for (p, nodes_visited) in enumerate(s.first_level_routes_nodes)
    val = s.first_level_routes_value[p]
    tour_cost = s.first_level_routes_cost[p] * val
    sol_cost += tour_cost
    println(" > 1st lev tour visiting \e[1m$nodes_visited\e[0m used $val time(s) (cost = $tour_cost).")
  end

  # Print data from second level
  ## Second level routes
  println("\e[1;34m Second level routes & customers visited.\e[0m")
  C = customers(data)
  customer_cov = zeros(Int, length(C))
  for (q, tour) in enumerate(s.second_level_routes_tour)
    nb_freighters_using_tour = s.second_level_routes_value[q]
    tour_cost = nb_freighters_using_tour * path_cost(data, tour)
    sol_cost += tour_cost
    for i in 2:(length(tour) - 1)
      customer_cov[tour[i]] += 1
    end
    println(" > 2nd lev tour \e[1m$tour\e[0m used $nb_freighters_using_tour time(s) (cost = $tour_cost).")
  end
  feasible_solution &= mapreduce(x -> (x == 1), &, customer_cov)
  !feasible_solution && @warn("At least one customer not covered.")

  # Link between two levels
  println("\e[1;34m Transshipments. \e[0m")
  S = satellites(data)
  for sat in S
    delivery = 0
    for (p, del) in s.items_delivered_to_sat[sat]
      delivery += del
    end
    shipment = s.items_shipped_from_sat[sat]
    feasible_solution &= (delivery >= shipment)
    !feasible_solution && @warn("Transshipmenets.")
    trans_cost = node_handling_cost(data, sat) * shipment
    sol_cost += (delivery > 0) ? trans_cost : 0
    sol_cost += (delivery > 0) ? node_cost(data, sat) : 0
    println(" > satellite \e[1m$sat\e[0m -- delivery = \e[1m$delivery\e[0m & shipment = \e[1m$shipment\e[0m -- cost = \e[1m$trans_cost\e[0m (h_cost = $(node_handling_cost(data, sat)) & setup_cost = $(node_cost(data, sat)) & cap = $(node_capacity(data, sat)))")
  end

  # Vehicles
  println("\e[1;34m Vehicles used. \e[0m")
  nb_urban_trucks = sum(s.first_level_routes_value)
  nb_city_freighters = sum(s.second_level_routes_value)
  feasible_solution &= (nb_urban_trucks <= veh_fleet_size(data, K))
  feasible_solution &= (nb_city_freighters <= veh_fleet_size(data, L))
  sol_cost += nb_urban_trucks * veh_cost(data, K) + nb_city_freighters * veh_cost(data, L)
  println(" > \e[1m $nb_urban_trucks \e[0m urban trucks used (out of $(veh_fleet_size(data, K))) -- setup_cost = $(veh_cost(data, K)).")
  println(" > \e[1m $nb_city_freighters \e[0m city freighters used (out of $(veh_fleet_size(data, L))) --  setup_cost = $(veh_cost(data, L)).")
  
  # Total cost
  println("Cost of the solution : \e[1m $sol_cost \e[0m")
  return feasible_solution
end
