abstract type Solution end

Solution(optimizer, data::Data2EVRP, v::Variables) = error("Solution extractor not implemented for Variables $(typeof(v)).")
set_variables_best_solution!(problem::Problem, s::Solution) = error("set_variables_best_solution! not implemented for type $(typeof(s)).")

const Path = Vector{Int} # [nodes]
const NodesVisited = Vector{Int} # [nodes]
const PathFreight = Tuple{Int, Double} # (terminal, quantity)

struct SolutionCapacitated <: Solution
    first_level_routes_nodes::Vector{NodesVisited}
    first_level_routes_value::Vector{Double}
    first_level_routes_cost::Vector{Double}
    second_level_routes_tour::Vector{Path}
    second_level_routes_value::Vector{Double}
    items_shipped_from_cdc::Dict{Int, Double} # d => nb_items
    items_delivered_to_sat::Dict{Int, Vector{PathFreight}} # s => [(p, freight)]
    items_shipped_from_sat::Dict{Int, Double} # s => demand
end

function path_cost(data::Data2EVRP, path::Path)::Double
    return sum([nodes_distance(data, path[i-1], path[i]) for i in 2:length(path)])
end

path_demand(data::Data2EVRP, path::Path) = mapreduce(x -> demand(data, x), +, path)

function register_1st_level_enumerated_routes!(optimizer, data::Data2EVRP, g::TierGraph, flr, flr_nb_trucks, flr_cost, P, λ)
    for (p, cost) in enumerate(P)
      val = round(get_value(optimizer, λ[0, p]))
      if val > 0
        push!(flr, nodes_visited_by_route(data, g, p))
        push!(flr_nb_trucks, val)
        push!(flr_cost, cost)
      end
    end
end

function extract_2nd_level_route(optimizer, g::TierGraph, starting_edge, depot, y, value)::Path
    path = Path()
    push!(path, depot)
    cur_node = (starting_edge[1] == depot) ? starting_edge[2] : starting_edge[1]
    cur_edge = starting_edge
    push!(path, cur_node)
    if value == 2 # If it's a two way trip (just 1 node visited)
        push!(path, depot)
        return path
    end
    while cur_node != path[1]
        found_next_node = false
        for e in incidents(g, cur_node)
            val = get_value(optimizer, y[depot, e])
            if val > 0 && e != cur_edge
                found_next_node = true
                cur_edge = e
                cur_node = (e[1] == cur_node) ? e[2] : e[1]
                push!(path, cur_node)
                break
            end
        end
        !found_next_node && error("Cannot create the path starting from $starting_edge. No edge found after $cur_edge.")
    end
    return path
end

sort(a,b) = (a < b) ? (a, b) : (b, a)
function register_2nd_level_routes!(optimizer, slr, slr_nb_freighters, g2l, y, S)
    for s in S
        forbidden_edges = Set{Tuple{Int64, Int64}}()
        for e in incidents(g2l[s], s)
            val = round(get_value(optimizer, y[s, e]))
            if val > 0 && e ∉ forbidden_edges
                path = extract_2nd_level_route(optimizer, g2l[s], e, s, y, val)
                push!(slr, path)
                if length(path) <= 3 # one node visited => arc used twice
                    val /= 2
                end
                push!(slr_nb_freighters, val)
                push!(forbidden_edges, sort(s, path[end - 1]))
            end
        end
    end
end

function demands_2nd_level_routes(data::Data2EVRP, slr, slr_nb_freighters, S)
    total_demand = Dict{Int, Float64}()
    for s in S
        total_demand[s] = 0
    end
    for (q, tour) in enumerate(slr_nb_freighters)
        total_demand[slr[q][1]] += path_demand(data, slr[q])
    end
    return total_demand
end

function Solution(optimizer, data::Data2EVRP, v::VarCapacitatedExtendedNoFreightFlow)
    S = satellites(data)
    D = cdc(data)
    K = veh_1l(data)[1]

    # First level routes (flr)
    flr_nodes_visited = Vector{NodesVisited}()
    flr_nb_trucks = Vector{Double}()
    flr_costs = Vector{Double}()
    register_1st_level_enumerated_routes!(optimizer, data, v.g1l, flr_nodes_visited, flr_nb_trucks, flr_costs, v.P, v.λ)

    # Second level routes
    println(" > Second level routes computing...")
    slr_routes = Vector{Path}()
    slr_nb_freighters = Vector{Double}()
    register_2nd_level_routes!(optimizer, slr_routes, slr_nb_freighters, v.g2l, v.y, S)
    nb_2lr = length(slr_routes)

    # Transfers at satellites
    ships_from_cdc = Dict{Int, Double}()
    deliveries_to_sats = Dict{Int, Vector{PathFreight}}()
    ships_from_sats = demands_2nd_level_routes(data, slr_routes, slr_nb_freighters, S)
    for d in D
        ships_from_cdc[d] = 0
    end
    for s in S
        deliveries_to_sats[s] = Vector{PathFreight}()
    end
    println(" > First level deliveries...")
    # 1st level delivery
    P = 1:length(flr_nodes_visited)
    model = Model(solver = CplexSolver())
    @variable(model, w[p in P, s in S] >= 0)
    @objective(model, Min, sum(w[p, s] for p in P, s in S))
    @constraint(model, knp[p in P], sum(w[p, s] for s in S) <= flr_nb_trucks[p] * veh_capacity(data, K))
    @constraint(model, wub[p in P, s in S], w[p, s] <= flr_nb_trucks[p] * veh_capacity(data, K) * (s ∈ flr_nodes_visited[p] ? 1 : 0))
    @constraint(model, dem[s in S], sum(w[p, s] for p in P) >= ships_from_sats[s])
    status = solve(model)
    (status != :Optimal) && error("Cannot compute 1st level paths deliveries.")

    for s in S, p in 1:length(P)
        items_shipped = getvalue(w[p, s])
        if items_shipped > 0
            push!(deliveries_to_sats[s], (p, items_shipped))
            ships_from_cdc[flr_nodes_visited[p][end]] += items_shipped
        end
    end
    return SolutionCapacitated(flr_nodes_visited, flr_nb_trucks, flr_costs, slr_routes, slr_nb_freighters, ships_from_cdc, deliveries_to_sats, ships_from_sats)
end
