import Base.show

abstract type GraphData end

struct Data2EVRP{T <: GraphData}
  instance::String
  nodes::DataFrame
  vehicles::DataFrame
  edges_weight_matrix
  nb_customers::Int
  nb_depots::Int
  nb_sats::Int
  nb_cdc::Int
  nb_items::Int
  graphdata::T
  set::ExperimentSet
end

function emptynodesframe4Data2EVRP()
  return DataFrame( Type = [], # cdc = D, depot = G, satellite = S, customer = C
                    X = [], # position
                    Y = [], # position
                    ST = [], # service time
                    A = [], # start of time window (expressed in time units)
                    B = [], # end of time window (expressed in time units)
                    D = [], # demand
                    H = [], # handling costs
                    Cap1 = [],  # Capacity 1st level
                    Cap2 = [], # Capacity 2nd level
                    Setup = [], # Setup cost
                    ItemCap = [] # Item capacity
                  )
end

function emptyvehiclesframe4Data2EVRP()
  return DataFrame( Name = [], # name of vehicle
                    Level = [], # level (1st or 2nd)
                    Cost = [], # cost
                    Capacity = [], # capacity
                    Amount = [] # amount of vehicles
                  )
end

function nodesid!(n::DataFrame)
  nb_cus = size(n[[occursin(r"C", e) for e in n.Type], :], 1)
  nb_dep = size(n[[occursin(r"G", e) for e in n.Type], :], 1)
  nb_sat = size(n[[occursin(r"S", e) for e in n.Type], :], 1)
  nb_cdc = size(n[[occursin(r"D", e) for e in n.Type], :], 1)
  # First two nodes are source and sink of the 2nd level graph
  # id_cus = 3  # (source and sink of the 2nd level are resp. nodes 1 & 2)
  # id_dep = nb_cus + 3
  # id_sat = nb_cus + nb_dep + 3
  # id_cdc = nb_cus + nb_dep + nb_sat + 3
  id_cus = 1
  id_dep = nb_cus + 1
  id_sat = nb_cus + nb_dep + 1
  id_cdc = nb_cus + nb_dep + nb_sat + 1
  n.IdGraph = [0 for i in 1:size(n, 1)]
  for i in 1:size(n, 1)
    if n[i, :Type] == "C"
      n[i, :IdGraph] = Int(id_cus)
      id_cus += 1
    elseif n[i, :Type] == "G"
      n[i, :IdGraph] = id_dep
      id_dep += 1
    elseif n[i, :Type] == "S"
      n[i, :IdGraph] = id_sat
      id_sat += 1
    elseif n[i, :Type] == "D"
      n[i, :IdGraph] = id_cdc
      id_cdc += 1
    else
      error("Unknown type of node : $(n[i, :Type]).")
    end
  end
  sort!(n, [order(:IdGraph)])
end

function create_depots!(n::DataFrame)
  sats = n[[occursin(r"S", e) for e in n.Type], :]
  nb_sats = size(sats, 1)
  coord = true
  for i in 1:nb_sats
    push!(n, ["G", sats[i, :X], sats[i, :Y], 0, NaN, NaN, 0, 0, sats[i, :Cap1], sats[i, :Cap2], sats[i, :MatrixCol]])
  end
  return nb_sats
end

function prepare_frames_2EVRP!(n::DataFrame, v::DataFrame)
  # Checking dataframe columns
  nb_dep = size(n[[occursin(r"G", e) for e in n.Type], :], 1)
  # Position of the node in the edge weights matrix
  n.MatrixCol = [i for i in 1:size(n, 1)]
  # If necessary, we create a depot for each satellite
  # if nb_dep == 0
  #   nb_dep = create_depots!(n)
  # end
  # Id of nodes and vehicles
  v.Id = [i for i in 1:size(v, 1)]
  nodesid!(n)
  return
end

function Data2ECVRP(n::DataFrame, v::DataFrame, ew, instancename::String)
  prepare_frames_2EVRP!(n, v)
  nb_cus = size(n[[occursin(r"C", e) for e in n.Type], :], 1)
  nb_dep = size(n[[occursin(r"G", e) for e in n.Type], :], 1)
  nb_sat = size(n[[occursin(r"S", e) for e in n.Type], :], 1)
  nb_cdc = size(n[[occursin(r"D", e) for e in n.Type], :], 1)
  if ew != nothing
    for i in 1:size(ew, 1)
      ew[i,i] = 0
    end
  end
  nb_nodes_l2 = nb_cus + nb_sat + nb_dep
  offset_graph_l1 = nb_cus + nb_dep
  nb_nodes_l1 = nb_sat + nb_cdc
  datagraph = Graph2ECVRP(nb_nodes_l2, nb_nodes_l1, offset_graph_l1)
  set = getexperimentset(instancename)
  return Data2EVRP(instancename, n, v, ew, nb_cus, nb_dep, nb_sat, nb_cdc, 0, datagraph, set)
end

Data2ECVRP(n::DataFrame, v::DataFrame, instancename::String) = Data2ECVRP(n, v, nothing, instancename)

## Getters

# MISC
instancename(d::Data2EVRP) = d.instance
bestsolutionvalue(d::Data2EVRP) = bestsolutionvalue(d.instance)
getub(d::Data2EVRP, w::Symbol) = ceil(ubvalue(d.instance, w) + 0.2)

# Nodes
nodes(d::Data2EVRP) = 1:size(d.nodes, 1)
nodesoftype(d::Data2EVRP, t) = Vector{Int64}(d.nodes[[t == e for e in d.nodes.Type], :][!, :IdGraph])
depots(d::Data2EVRP) = nodesoftype(d, "G")
cdc(d::Data2EVRP) = nodesoftype(d, "D")
satellites(d::Data2EVRP) = nodesoftype(d, "S")
customers(d::Data2EVRP) = nodesoftype(d, "C")

function vehicleidsoflevel(d::Data2EVRP, l)
  return Vector(d.vehicles[[l == level for level in d.vehicles.Level], :][!, :Id])
end

veh_1l(d::Data2EVRP) = vehicleidsoflevel(d, 1)
veh_2l(d::Data2EVRP) = vehicleidsoflevel(d, 2)
veh_capacity(d::Data2EVRP, vid) = d.vehicles.Capacity[vid]
veh_cost(d::Data2EVRP, vid) = d.vehicles.Cost[vid]
veh_fleet_size(d::Data2EVRP, vid) = d.vehicles.Amount[vid]
veh_type(d::Data2EVRP, vid) = d.vehicles.Type[vid]

node_veh1l_parksize(d::Data2EVRP, n) = d.nodes.Cap1[n]
node_veh2l_parksize(d::Data2EVRP, n) = d.nodes.Cap2[n]
node_handling_cost(d::Data2EVRP, n) = d.nodes.H[n]
node_cost(d::Data2EVRP, n) = d.nodes.Setup[n]
node_capacity(d::Data2EVRP, n) = d.nodes.ItemCap[n]

x_coord(d::Data2EVRP, n) = d.nodes.X[n]
y_coord(d::Data2EVRP, n) = d.nodes.Y[n]
demand(d::Data2EVRP, c) = d.nodes.D[c]
totaldemand(d::Data2EVRP) = sum(d.nodes.D)

nbcustomers(d::Data2EVRP) = d.nb_customers
nbdepots(d::Data2EVRP) = d.nb_depots
nbcdc(d::Data2EVRP) = d.nb_cdc
nbsatellites(d::Data2EVRP) = d.nb_sats

function lb_nb_sats_required(data::Data2EVRP)::Int
  sat_capacities = [node_capacity(data, s) for s in satellites(data)]
  sort!(sat_capacities, rev = true)
  capacity_available = 0
  lbnbsats = 0
  while totaldemand(data) > capacity_available
    capacity_available += sat_capacities[lbnbsats + 1]
    lbnbsats += 1
  end
  return lbnbsats
end

function lb_nb_cdc_required(data::Data2EVRP)::Int
  cdc_capacities = [node_capacity(data, d) for d in cdc(data)]
  sort!(cdc_capacities, rev = true)
  capacity_available = 0
  lbnbcdc = 0
  while totaldemand(data) > capacity_available
    capacity_available += cdc_capacities[lbnbcdc + 1]
    lbnbcdc += 1
  end
  return lbnbcdc
end

function is_second_level_edge(d::Data2EVRP, n1::Int, n2::Int)::Bool
  return n1 <= nbcustomers(d) || n2 <= nbcustomers(d)
end

function is_first_level_edge(d::Data2EVRP, n1::Int, n2::Int)::Bool
  return !secondlevel_edge(d, n1, n2)
end

function nodes_distance(d::Data2EVRP, n1::Int, n2::Int)::Double
  return nodes_distance(d, n1, n2, typeof(d.set))
end

function euclidean_dist(d::Data2EVRP, n1::Int, n2::Int)::Double
  x_d = (x_coord(d, n1) - x_coord(d, n2))^2
  y_d = (y_coord(d, n1) - y_coord(d, n2))^2
  return √(x_d + y_d) 
end

function nodes_distance(d::Data2EVRP, n1::Int, n2::Int, ::Type{<: ExperimentSet})::Double
  return euclidean_dist(d, n1, n2)
end

function nodes_distance(d::Data2EVRP, n1::Int, n2::Int, ::Type{<: Set1})::Double
  n1_col = d.nodes[n1, :MatrixCol]
  n2_col = d.nodes[n2, :MatrixCol]
  return d.edges_weight_matrix[n1_col, n2_col]
end

function distsubmatrix(d::Data2EVRP, terminals::Array)
  return distsubmatrix(d, terminals, terminals)
end

function distsubmatrix(d::Data2EVRP, terminals_cols::Array, terminals_rows::Array)
  submatrix = Matrix{Float64}(length(terminals_rows), length(terminals_cols))
  for (i, ti) in enumerate(terminals_rows), (j, tj) in enumerate(terminals_cols)
    submatrix[i, j] = dist(d, ti, tj)
  end
  return submatrix
end

function cost(d::Data2EVRP, e)
  e = Tuple(e)
  return nodes_distance(d, e[1], e[2])
end

route_cost(d::Data2EVRP, route) = sum([cost(d, e) for e in route])

function search_node(data::Data2EVRP, x::Integer, y::Integer, d::Integer, t::String)
  set1 = Vector{Int64}(data.nodes[[t == e for e in data.nodes.Type], :][:IdGraph])
  set2 = Vector{Int64}(data.nodes[[x == e for e in data.nodes.X], :][:IdGraph])
  set3 = Vector{Int64}(data.nodes[[y == e for e in data.nodes.Y], :][:IdGraph])
  set4 = Vector{Int64}(data.nodes[[d == e for e in data.nodes.D], :][:IdGraph])
  intersect = set1 ∩ set2 ∩ set3 ∩ set4
  if length(intersect) != 1
    println("********************")
    println("********************")
    error("$(length(intersect)) node(s) found for search with args : x = $x, y = $y, d = $d & t = $t.")
  end
  return intersect[1]
end

function print_messages_about_instance(data::Data2EVRP)
  nb_cust_with_demand_zero = count([demand(data, c) for c in customers(data)] .== 0)
  if nb_cust_with_demand_zero > 0
    @warn("$nb_cust_with_demand_zero customers have a demand equals 0.")
  end
end

function show(io::IO, data::Data2EVRP)
  K = veh_1l(data)
  L = veh_2l(data)
  println(io, "Instance \e[1m $(data.instance) \e[0m")
  println(io, "┋  > nb of city distribution centers : \e[1m $(length(cdc(data))) \e[0m")
  println(io, "┋  > nb of satellites : \e[1m $(length(satellites(data))) \e[0m ")
  println(io, "┋  > nb of customers : \e[1m $(length(customers(data))) \e[0m ")
  println(io, "┋  > nb of depots : \e[1m $(length(depots(data))) \e[0m ")
  println(io, "┋  > nb of 1st level vehicles : \e[1m $(veh_fleet_size(data, K))\e[0m  \t capacity : \e[1m$(veh_capacity(data, K))\e[0m")
  println(io, "┋  > nb of 2nd level vehicles : \e[1m $(veh_fleet_size(data, L))\e[0m  \t capacity : \e[1m$(veh_capacity(data, L))\e[0m")
  #println(io, "┋ Best solution found in literature : \e[1m $(bestsolutionvalue(data)) \e[0m  -- UB provided : \e[1m $(bestub(data)) \e[0m")
end
