function subsets_of_satellites(data::Data2EVRP)::Vector{Vector{Int}}
    S = satellites(data)
    subsets = Vector{Vector{Int}}()
    number_of_subsets = 2^length(S) - 1
    for i in 1:number_of_subsets
        subset = Vector{Int}()
        ibin = string(i, base = 2)
        lbin = length(ibin)
        for j in 1:lbin
        if ibin[j] == '1'
            push!(subset, S[lbin + 1 - j])
        end
        end
        push!(subsets, subset)
    end
    return subsets
end

function build_model(data::Data2EVRP, param::Dict)
  model = VrpModel()
  ## Sets
  D = cdc(data)
  S = satellites(data)
  C = customers(data)

  g1l = graph_1l(data)
  g2lmain = graph_2l(data, S)
  g2l = Dict(s => graph_2l(data, s) for s in S)
  V1 = vertice(g1l)
  V2 = vertice(g2lmain)
  E1 = edges(g1l)

  E2 = Dict(s => edges(g2l[s]) for s in S)
  
  K = veh_1l(data)[1] # Only one type of vehicles
  L = veh_2l(data)[1]

  Q1 = veh_capacity(data, K)
  Q2 = veh_capacity(data, L)

  p1l_costs = heldkarp_tsp_enum(data, g1l)
  P = 1:length(p1l_costs)

  u_var_subset_size = 5

  SatellitesSubSets = subsets_of_satellites(data)
  SSS = 1:length(SatellitesSubSets)

  nbmaxurbantrucks = veh_fleet_size(data, K)
  demands = [demand(data, i) for i in union(C, S)]

  # -- VARIABLES d
  # --- First level variables
  @variable(model.formulation, 0 <= λ[0, p in P] <= veh_fleet_size(data, K), Int)
  @variable(model.formulation, t[0, i in [1 2]] >= 0, Int)
  # --- Second level variables
  @variable(model.formulation, 0 <= y[s in S, e in E2[s]] <= 2, Int)
  # --- Number of urban trucks visiting a subset of satellites
  @variable(model.formulation, u[0, sss in SSS] >= 0, Int)
  # -- Demand variable at satellite (limit number of var of subproblem in cstr)
  @variable(model.formulation, d[0, s in S] >= 0)

  # -- OBJECTIVE FUNCTION
  @objective(model.formulation, Min, sum(p1l_costs[p] * λ[0, p] for p in P) # 1st level
    + sum(cost(data, e) * y[s, e] for s in S, e in E2[s]) # 2nd level
    + sum(node_handling_cost(data, s) * d[0,s] for s in S) # Handling costs
  )

  # -- CONSTRAINTS
  ## Seconds level routes visit all customers
  @constraint(model.formulation, cust_cov[0, c in C],
    sum(y[s, e] for s in S, e in incidents(g2l[s], c)) == 2
  )

  ## Number of 1st level vehicles & second level vehicles used (set var t)
  @constraint(model.formulation, nb_1stveh_used[0],
    t[0,1] == sum(λ[0,p] for p in P)
  )
  @constraint(model.formulation, nb_2ndveh_used[0],
    t[0,2] == sum(0.5 * y[s, e] for s in S, e in incidents(g2l[s], s))
  )

  ## Number of first level vehicles
  ### Lower bound
  @constraint(model.formulation, nb_1stveh_lb[0],
    t[0,1] >= ceil(totaldemand(data) / Q1)
  )
  
  ### Upper bounds
  @constraint(model.formulation, nb_1stveh_ub[0],
    t[0,1] <= veh_fleet_size(data, K)
  )

  ## Number of second level vehicles
  ### Upper bound
  @constraint(model.formulation, nb_2ndveh_ub[0],
    t[0,2] <= veh_fleet_size(data, L)
  )
  ### Lower bound
  @constraint(model.formulation, nb_2ndveh_lb[0],
    t[0,2] >= ceil(totaldemand(data) / Q2)
  )

  @constraint(model.formulation, u_val[0, sss in SSS],
    u[0,sss] == t[0,1] - sum(λ[0,p] for p in P[.!routes_visiting(data, sss)])
  )

  @constraint(model.formulation, twolev[0, sss in SSS],
    sum(d[0,s] for s in SatellitesSubSets[sss]) <= Q1 * u[0,sss]
  )

  ## Compute the demand of customers at satellite
  @constraint(model.formulation, dcust_at_sat[0, s in S],
    d[0,s] == sum(0.5 * demand(data, c) * y[s, e] for c in C, e in incidents(g2l[s], c))
  )
  
  function rcsp_cvrp_sat_solver(s)
    ## The startingpoint & the sink is the satellite
    nb_C = length(C)
    source = sink = s
    nodes = vcat(C, s)

    fleet_size = veh_fleet_size(data, L)
    park_size = node_veh2l_parksize(data, s)
    max_mult = Int(minimum([fleet_size, park_size]))

    network = VrpGraph(model, nodes, source, sink, (0, max_mult))

    cap_res = add_resource!(network, main = true)

    # RCSP functions
    function create_arc!(queue, tail, variables)
        cap_consumption = 0.5 * (demand(data, queue) + demand(data, tail))
        arc_id = add_arc!(network, queue, tail)
        for var in variables
            add_arc_var_mapping!(network, arc_id, var)
        end
        set_arc_consumption!(network, arc_id, cap_res, cap_consumption)
    end

    function create_arc!(queue, tail)
        edge = sort(queue, tail)
        create_arc!(queue, tail, [y[s, edge]])
    end

    function create_bipartite!(A, B)
        for a in A, b in B
            create_arc!(a, b)
        end
    end

    function create_complete!(A)
        for a in A, b in A
            if a != b 
                create_arc!(a, b) 
                create_arc!(b, a)
            end
        end
    end

    create_bipartite!([source], C)
    create_bipartite!(C, [sink])
    create_complete!(C)

    for i in nodes
        set_resource_bounds!(network, i, cap_res, 0, veh_capacity(data, L))
    end
    return network
  end

  graphs = []
  for s in S
    network = rcsp_cvrp_sat_solver(s)
    add_graph!(model, network)
    push!(graphs, network)
  end

  set_vertex_packing_sets!(model, [[(G,c) for G in graphs] for c in C])
  add_capacity_cut_separator!(model, [([(G,c) for G in graphs], Float64(demand(data, c))) for c in C], Float64(Q2))
  for G in graphs
    define_elementarity_sets_distance_matrix!(model, G, [[cost(data, (i, j)) for i in C] for j in C])
  end

  enable_rank1_cuts!(model)


  # Branching priorities of variables
  set_branching_priority!(model, "u", 2)
  set_branching_priority!(model, "t", 1)
  set_branching_priority!(model, "λ", 1)
  set_branching_priority!(model, "y", 1)

  @expression(model.formulation, vehNum[s in S], sum(0.5 * y[s, e] for e in incidents(g2l[s], s)))
  set_branching_priority!(model, vehNum, "vehNum", 1)

  @expression(model.formulation, assign[s in S, c in C], sum(0.5 * y[s, e] for e in incidents(g2l[s], c)))
  set_branching_priority!(model, assign, "assign", 1)

  ### --------------------------------------------------
  ### Valid inequalities
  ### --------------------------------------------------
  sep_metrics = Vector{SeparationMetrics}()

  ssi_met = SeparationMetrics("SSI")
  osi_met = SeparationMetrics("OSI")

  ssi_vic = ViolatedInequalities{SatellitesSupplyInequality}(ssi_met, 150, true)
  osi_vic = ViolatedInequalities{OpenSatelliteInequality}(osi_met, 50, false)

  push!(sep_metrics, ssi_met)
  push!(sep_metrics, osi_met)

  ### --------------------------------------------------
  ### Separation callback
  ### --------------------------------------------------
  
  λ_col_offset = λ[0,1].col - 1
  function validineq_callback()::Cvoid
    n = length(C) + length(S)

    y_vals = zeros(Float64, n, n, length(S))
    offset = S[1] - 1
    for s in S, e in E2[s]
      y_vals[e[1], e[2], s - offset] = get_value(model.optimizer, y[s, e])
      y_vals[e[2], e[1], s - offset] = get_value(model.optimizer, y[s, e])
    end

    cust_demands = Double[demand(data, c) for c in C]
    d_vals = Double[get_value(model.optimizer, d[0, s]) for s in S]
    λ_vals = Double[get_value(model.optimizer, λ[0, p]) for p in P]
    u_vals_sat = Double[get_value(model.optimizer, u[0, 2^(s - S[1])]) for s in S]
    
    SSI_separation(model.optimizer, λ_vals, y_vals, cust_demands, d_vals)
    OSI_separation(model.optimizer, u_vals_sat, y_vals)
    return
  end

  function SSI_separation(optimizer, λ_vals, y_vals, cust_demands, d_vals)::Cvoid
    separate_ssi!(ssi_vic, λ_vals, y_vals, C, S, Q1, Q2, g2lmain, cust_demands, 
        d_vals)
    separation_done!(ssi_vic)
    println("SSI : $(length(ssi_vic)) found.")
    for id in ssi_vic
      ssi_ineq = getinequality(ssi_vic, id)
      vars = []
      coeffs = []
      if count_ones(ssi_ineq.satsubsetid) <= u_var_subset_size
        push!(vars, u[0,ssi_ineq.satsubsetid])
        push!(coeffs, -2.0 * ssi_ineq.m_value)
      else
        push!(vars, t[0,1])
        push!(coeffs, -2.0 * ssi_ineq.m_value)
        for p in P[.!routes_visiting(data, ssi_ineq.satsubsetid)]
          push!(vars, λ[0,p])
          push!(coeffs, +2.0 * ssi_ineq.m_value)
        end
      end
      sats = setdiff(S, SatellitesSubSets[ssi_ineq.satsubsetid])
      for e in getedges(ssi_ineq), s in sats
        e[2] ∈ S && e[2] != s && continue
        e[1] ∈ S && e[2] ∈ S && continue
        push!(vars, y[s, e])
        push!(coeffs, 1.0)
      end
      add_dynamic_constr!(optimizer, vars, coeffs, >=, ssi_ineq.rhs_value, "validineq")
    end
    return
  end

  function OSI_separation(optimizer, u_vals_sat, y_vals)
    sep_time = @elapsed separate_osi!(osi_vic, S, C, u_vals_sat, y_vals)
    separation_done!(osi_vic, sep_time)
    println("OSI : $(length(osi_vic)) found.")
    for id in osi_vic
      osi_ineq = getinequality(osi_vic, id)
      vars = []
      coeffs = []
      push!(vars, u[0,osi_ineq.var_u_id])
      push!(coeffs, -2)
      for e in incidents(g2l[osi_ineq.satellite], osi_ineq.customer)
        push!(vars,  y[osi_ineq.satellite, e])
        push!(coeffs, 1)
      end
      add_dynamic_constr!(optimizer, vars, coeffs, <=, 0, "validineq")
    end
    return
  end
  add_cut_callback!(model, validineq_callback, "validineq")

  return model, Variables(p1l_costs, g1l, g2l, λ, y), sep_metrics
end
