# Author : Guillaume Marques
# Description : Containers for violated inequalities

abstract type ViolatedInequality end

function duplicated(a::ViolatedInequality, b::ViolatedInequality)::Bool
  error("Method not implemented for ", typeof(a), ".")
  return false
end

function violationvalue(a::ViolatedInequality)::Double
  error("Method not implemented for ", typeof(a), ".")
  return 0.0
end

mutable struct ViolatedInequalities{T <: ViolatedInequality}
  nb_max_ineq::Int
  ineqs_added::Int
  use_buffer::Bool
  violated_inequalities::Array{T, 1}
  buffer::Array{T, 1}
  metrics::SeparationMetrics
  separation_done::Bool
  function ViolatedInequalities{T}(metrics::SeparationMetrics, m::Int, b::Bool
      ) where T <: ViolatedInequality
    return new(m, 0, b, Vector{T}(), Vector{T}(), metrics, false)
  end
  function ViolatedInequalities{T}(metrics::SeparationMetrics
      ) where T <: ViolatedInequality
    return new(1000000, 0, false, Vector{T}(), Vector{T}(), metrics, false)
  end
end

import Base.length
function length(vic::ViolatedInequalities{T})::Int where T <: ViolatedInequality
  remove_duplicates!(vic.violated_inequalities)
  return length(vic.violated_inequalities)
end

function buf_length(vic::ViolatedInequalities{T})::Int where 
    T <: ViolatedInequality
  return length(vic.buffer)
end

function buf_maxviol(vic::ViolatedInequalities{T})::Double where 
    T <: ViolatedInequality
  (buf_length(vic) == 0) && return 0.0
  violations = map(violationvalue, vic.buffer)
  return maximum(violations)
end

function add_violated_inequality!(vic::ViolatedInequalities{T}, i::T
    )::Cvoid where T <: ViolatedInequality
  vic.separation_done && error("Write access forbidden while separation is done.")
  if vic.use_buffer
    push!(vic.buffer, i)
  else
    push!(vic.violated_inequalities, i)
  end
  return
end

function remove_duplicates!(cont::Array{T, 1})::Cvoid where 
    T <: ViolatedInequality
  (length(cont) == 0) && return
  sort!(cont, by = c -> violationvalue(c), rev = true)
  rm_ids = Vector{Int}()
  cur_ineq = cont[1]
  c_from = 0
  c_to = 0 
  size_cont = length(cont)
  while c_to < size_cont
    # Update the comparison window (using violation value)
    c_from = c_to + 1
    c_to = c_from
    while (c_to < size_cont) && 
        (violationvalue(cont[c_from]) == violationvalue(cont[c_to]))
      c_to += 1
    end
    # Store ids of inequalities to remove
    for i in c_from:c_to, j in i:c_to
      if i != j
        if duplicated(cont[i], cont[j])
          push!(rm_ids, i)
          break
        end
      end
    end
  end
  sort!(rm_ids)
  deleteat!(cont, rm_ids)
  return
end

function register_metrics!(vic::ViolatedInequalities{T}, cont::Array{T, 1},
    sep_time::Double)::Cvoid where T <: ViolatedInequality
  nb_ineqs = min(length(cont), vic.nb_max_ineq)
  if nb_ineqs > 0
    viol_vals = map(violationvalue, cont[1:nb_ineqs])
    avgviol = mean(viol_vals)
    maxviol = maximum(viol_vals)
    register_metrics!(vic.metrics, sep_time, nb_ineqs, avgviol, maxviol)
  else
    register_metrics!(vic.metrics, sep_time, 0, 0.0, 0.0)
  end
  return
end

function flush_buffer!(vic::ViolatedInequalities{T}, sep_time::Double
    )::Cvoid where T <: ViolatedInequality
  if !vic.use_buffer 
    @warn("Buffer not used for ", T)
    return
  end
  (length(vic.buffer) == 0) && return
  remove_duplicates!(vic.buffer)
  sort!(vic.buffer, by = vi -> violationvalue(vi), rev = true)
  register_metrics!(vic, vic.buffer, sep_time)
  lastid = min(vic.nb_max_ineq, length(vic.buffer))
  for i in 1:lastid
    push!(vic.violated_inequalities, vic.buffer[i])
  end
  empty!(vic.buffer)
  return
end

function separation_done!(vic::ViolatedInequalities{T}
  )::Cvoid where T <: ViolatedInequality
(length(vic.buffer) > 0) && error("Buffer is not empty. Did you flush it?")
!vic.use_buffer && error("You must use separation_done!(vic, sep_time).")
vic.separation_done = true
return
end

function separation_done!(vic::ViolatedInequalities{T}, sep_time::Double
    )::Cvoid where T <: ViolatedInequality
  (length(vic.buffer) > 0) && error("Buffer is not empty. Did you flush it?")
  register_metrics!(vic, vic.violated_inequalities, sep_time)
  vic.separation_done = true
return
end

function Base.iterate(vic::ViolatedInequalities{T}) where {T<:ViolatedInequality}
    !vic.separation_done && error("Did you notify the container that separation is done?")
    if length(vic.violated_inequalities) == 0 
        vic.separation_done = false
        return nothing
    end
    remove_duplicates!(vic.violated_inequalities)
    sort!(vic.violated_inequalities, by = vi -> violationvalue(vi), rev = true)
    return (1 + vic.ineqs_added, 1)
end

function Base.iterate(vic::ViolatedInequalities{T}, state::Int
    ) where T <: ViolatedInequality
    nb_ineqs_added =  min(vic.nb_max_ineq, length(vic.violated_inequalities))
    if state > nb_ineqs_added
        empty!(vic.violated_inequalities)
        vic.ineqs_added += nb_ineqs_added
        vic.separation_done = false
        return nothing
    end
    return (state + vic.ineqs_added, state + 1)
end

function getinequality(vic::ViolatedInequalities{T}, id::Int
    )::T where T <: ViolatedInequality
  !vic.separation_done && error("Read access forbidden while separation is not done.")
  return vic.violated_inequalities[id - vic.ineqs_added]
end
